(*
 *
 * Copyright (C) 2002 Aurelien Moreau, Antoni Gutsens Diaz, Benjamin Leveque,
 * 	 Damien Pous, David Roger, Lionel Vaux, Mathieu Martin, Nicolas Bernard,
 * 	 Samuel Thibault, Simon Capern, Thomas Gazagnaire
 * 
 * This file is part of Rigobot.
 * 
 * Rigobot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Ribogot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ribogot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)
(*
 * faudra trouver un super entte pour mettre sur nos fichiers
 *
 *
 *)

(* Cell : dfinition de la classe abstraite 'cell' 
 * 
 * le paramtre de type ('obj) traduit le fait qu'une case contient des 
 * objets de type 'obj 
 * 
 *)


open Position

(* exception lance lorsqu'un objet tente d'aller sur une case occuppe *)
exception Busy

(* execption lance lorsque l'on accde  certaines mthode de la case 
 * spciale "invalid_cell". (cf plus bas...)
 *)
exception Invalid


(********** type de la classe cell, utilise pour dfinir gl_cell_t **********)
class type ['obj] cell_t = 
object
  (* position de la case *)
  method x: int
  method y: int
  method pos: pos_t

  (* liste des objets tant sur la case *)
  method objets: 'obj list

  (* retrait d'un objet *)
  method rem: 'obj -> unit

  (* retrait des objets "inutiles" *)
  method clean: unit

  (* dplacement d'un objet vers self 
   * si cet objet n'a pas le droit, l'exception Busy doit tre leve
   *)
  method accept: 'obj -> dir_t -> unit

  (* affichage (gnrique) *)
  method draw: unit

  (* affichage (mode texte) *)
  method print: unit
end


(********** la classe case proprement dite **********)
class virtual ['obj] cell ((x,y): pos_t) =
object (self)
  (* coercion en un cell_t*)
  method coerce: 'obj cell_t = (self :> 'obj cell_t)

  (* position de la case *)
  val p = ((x,y): pos_t)
  method pos = p
  method x = x
  method y = y

  (* liste des objets tant sur la case *)
  val mutable list: 'obj list = []
  method objets = list
  method private iter f = List.iter f list
  method private fold_left f v = List.fold_left f v list


  (* ajout, retrait d'objets *)
  method private add o = list <- o::list
  method rem o = 
    (* on ne fait rien si l'objet n'est pas trouv *)
    let rec aux accu = function
      | x::q when x == o -> List.rev_append accu q
      | x::q -> aux (x::accu) q
      | []   -> accu
    in
      list <- aux [] list;

  (* retrait des objets "inutiles" *)
  method clean = 
    list <- List.filter (fun o -> o#keep) list

  (* ajout d'un objet, s'il en a le droit 
   * le cas chant, l'exception Busy est leve
   *)
  method accept =
    let rec aux l obj dir = match l with
      | x::q -> 
	  ignore (x#is_valid obj dir  || raise Busy);
	  aux q obj dir
      | []   -> 
	  obj#cell#rem obj;
	  obj#set_cell self#coerce;
	  self#add obj 
    in
      aux list
     

  (* affichage (gnrique) *)
  method virtual draw: unit


  (* sortie  l'cran *)
  method print = match list with
    | [] -> print_char ' '
    | o::_ -> o#print
    
end



(********** classe invalid_cell **********)

(* cette classe permet de crer une case "invalide"    
 * vers laquelle le monde peut se tourner lorsque l'on accde 
 *  une case hors de son tableau.
 * elle permettent aussi de crer un mur fictif
 * (la mthode 'is_valid' levant tjs l'exception)
 *)
class ['obj] invalid_cell = 
object (self)
  (* on hrite de cell afin de ne pas tre embt si de nouvelles choses 
   * sont ajoutes  cell_t...
   * il faut par contre faire attention de bien lever des exception
   * lorsque c'est ncessaire !
   *)
  inherit ['obj] cell (-1,-1)

  (* un simple raccourci *)
  method private err = raise Invalid

(* doit-on lever une exception ou renvoyer [] ? *)
(*   method objets = self#err *)

  (* retrait d'objet : impossible ! *)
  method rem o = self#err

  (* ajout : la case est inaccessible ! *)
  method accept obj dir = raise Busy

  (* affichage (gnrique) *)
  (* ici aussi, on pourrait lever l'exception... mais bon :-) *)
  method draw = 
    print_string "Warning, trying to render the invalid cell !";
    print_newline()

  (* affichage (mode texte) *)
  method print = print_char '#'

end



(********** obtention des objets d'une liste de cases **********)
let scan: 'obj cell_t list -> 'obj list = 
  fun l -> List.fold_left (fun acc cell -> List.rev_append cell#objets acc) [] l

