(*
 *
 * Copyright (C) 2002 Aurelien Moreau, Antoni Gutsens Diaz, Benjamin Leveque,
 * 	 Damien Pous, David Roger, Lionel Vaux, Mathieu Martin, Nicolas Bernard,
 * 	 Samuel Thibault, Simon Capern, Thomas Gazagnaire
 * 
 * This file is part of Rigobot.
 * 
 * Rigobot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Ribogot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ribogot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)
(*
 * faudra trouver un super entte pour mettre sur nos fichiers
 *
 *
 *)

(* Objet : dfinition de la classe abstraite 'objet' 
 * 
 * cette classe est paramtre par trois variables de type :
 * 'world est le type du monde dans lequel l'objet se trouve
 * 'kind  permet de dfinir la mthode "kind" qui rend le genre 
 *    d'un objet (champignon, robot...)
 * 'coerce indique que l'objet peut tre coerc (rduit) en un objet
 *    type 'coerce (utile pour crer des listes de {robots, champignons, ...}
 * 
 * 
 * 
 * Gestion des vnements/actions :
 *   trois mthodes servent  grer les vnements :
 *     -  [generate] pour les objets qui veulent en gnrer 
 *     -  [notify]   pour ceux qui veulent en tenir compte
 *     -  [react]    pour ceux qui veulent agir lors d'un tour
 * 
 *   [generate] est appele au dbut de chaque tour. Cette mthode peut
 *     gnrer des vnements  pour objet o en appelant o#declare.
 *     si elle veut gnrer un vnement pour tous les objets, il suffit 
 *     qu'elle le mette dans la liste quelle renvoie
 *   [notify]   sert  dclarer  un objet qu'un vnement a t gnr
 *     celui-ci peut elors en faire ce qu'il veut (ie. l'ignorer ou le retenir
 *   [react]    est appele par le monde  chaque tour, une fois que tous les 
 *     objets ont dclar ce qu'ils voulaient.
 * 
 * Dans un soucis d'efficacit (au moins thorique),
 * un objet implmentant vraiment ces mthodes doit le dclarer
 *  l'aide du paramtre 'options' de la classe
 * (cf. l'initializer)
 *)


open Position
open Cell

(* ce module permet de numroter les objets *)
module Id = Numbered.Create(Numbered.Null)


(********** type de la classe, utilis par gl_objet **********)
class type ['kind, 'world, 'coerce] objet_t = 
object

  (* chaque objet est identifi de manire unique *)
  method id: int

  (* 'genre' de l'objet*)
  method kind: 'kind

  (* case sur laquelle est l'objet *)
  method cell: 'coerce cell_t
  method set_cell: 'coerce cell_t -> unit

  (* monde dans lequel est l'objet *)
  method world: 'world

  (* position dans le monde (=cell#pos) *)
  method pos: pos_t
    
  (* zone "d'influence" dans le monde *)
  method zone: pos_t list
    
  (* direction (si elle est dfinie pour l'objet) *)
  method dir: dir_t

  (* tat actuel en termes de saut *)    
  method jump_state: Jumping.state_t
   
  (* dplacement vers une nouvelle position *)
  method move: dir_t -> unit

  (* vnements globaux  gnrer pour le top courant 
   * cette mthode doit aussi en profiter pour gnrer les vnements 
   * locaux ncessaires, via declare
   *)
  method generate: Events.event list

  (* dclaration d'un vnement pour self *)
  method notify: Events.event -> unit

  (* raction *)
  method react: unit

  (* un objet donn peut-il traverser l'objet(self) dans une 
   * direction donne ? 
   *)
  method is_valid: 'coerce -> dir_t -> bool

  (* l'objet est-il "utile", doit-il tre conserv sur la carte ? 
   * un objet "inutile " est par exemple, un lvateur d'altitude :
   * il ne sert qu' l'initialisation...
   *)
  method keep: bool

  (* affichage (mode texte) *)
  method print: unit
end



(********** la classe de base objet ************)
class virtual ['kind, 'world, 'coerce] objet 
  (options: [`Generate | `Listen | `React] list) 
  (world: 'world)
  pos 
  = 
object (self)

  (* fournit la mthode id: int, qui numrote les objets de faon unique *)
  inherit Id.numbered 

  (* fermeture du type de l'objet *)
  (* (cette mthode n'est dfinie que dans le monde 3d, o l'on connait *)
  (* tout ce qui est ncessaire) *)
  method virtual coerce: 'coerce
    
  (* case sur laquelle est l'objet *)
  val mutable cell: 'coerce cell_t = world#cell pos
  method cell = cell
  method set_cell cell' = cell <- cell'

  (* monde dans lequel est l'objet*)
  method world = world

  (* position dans le monde (par le bias de la case)*)
  method pos = cell#pos

  (* zone "d'influence" dans le monde : par dfaut, notre case *)
  method zone = [self#pos]
    
  (* direction : par dfaut, on n'en a pas *)
  method dir = null_dir

  (* tat actuel en termes de saut : au sol par dfaut *)    
  method jump_state: Jumping.state_t = `Normal

  (* dplacement dans le monde *)
  method move = world#deplace self#coerce

  (* un objet donn peut-il traverser l'objet(self) dans une 
   * direction donne ? 
   * par dfaut, un objet peut toujours tre travers
   *)
  method is_valid (obj: 'coerce) (dir: dir_t) = true

  (* vnements  gnrer pour le top courant 
   * par dfaut, aucun.
   *)
  method generate: Events.event list = []

  (* on nous dclare un vnement, 
   * par dfaut, on s'en fiche.
   *)
  method notify (e: Events.event) = ()

  (* action lors d'un tour,
   * par dfaut, rien.
   *)
  method react = ()

  (* par dfaut, tout objet est utile *)
  method keep = true

  (* affichage (mode texte) *)
  method print = print_char 'o'

  initializer
    (* on doit imprativement s'enregistrer sur sa case *)
    cell#accept self#coerce null_dir;
    (* on enregistre aussi les traitants ncessaires *)
    if List.mem `Generate options then world#register_generator (fun () -> self#generate);
    if List.mem `Listen   options then world#register_listener  (self#notify);
    if List.mem `React    options then world#register_reacter   (fun () -> self#react);
    
end



