(*
 *
 * Copyright (C) 2002 Aurelien Moreau, Antoni Gutsens Diaz, Benjamin Leveque,
 * 	 Damien Pous, David Roger, Lionel Vaux, Mathieu Martin, Nicolas Bernard,
 * 	 Samuel Thibault, Simon Capern, Thomas Gazagnaire
 * 
 * This file is part of Rigobot.
 * 
 * Rigobot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Ribogot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ribogot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)
(*
 * faudra trouver un super entte pour mettre sur nos fichiers
 *
 *
 *)

(* World : dfinition de la classe abstraite 'world' 
 * 
 * le paramtre de type ('obj) traduit le fait qu'un monde contient des 
 * objets de type 'obj
 * 
 *)


open Position
open Cell


(* type rendu par le monde aprs chaque coup *)
type status_t = Won | Lost | Continue




(* type de la classe world, utilise pour dfinir gl_world_t *)
class type ['obj] world_t = 
object

  (* obtention d'une case du terrain *)
  method cell: pos_t -> 'obj cell_t

  (* obtention des objets dans un certain rayon *)
  method cell_list: pos_t list -> 'obj cell_t list

  (* jouer un tour *)
  method next: status_t

  (* affichage (mode texte) *)
  method print: unit

  (* enregistrement d'un nouveau gnrateur d'vnements *)
  method register_generator: (unit -> Events.event list) -> unit

  (* enregistrement d'un nouvel couteur d'vnements *)
  method register_listener: (Events.event -> unit) -> unit

  (* enregistrement d'un nouvel actionneur *)
  method register_reacter: (unit -> unit) -> unit

  (* retrait des objets "inutiles" *)
  method clean: unit

  (* dplacement d'un objet 
   * cette mthode doit lever l'exception Cell.Busy si la case est occupe
   *)
  method deplace: 'obj -> dir_t -> unit

  (*  appeler lors d'un tour pour indiquer que l'on a gagn ou perdu *)
  method won: unit
  method lost: unit

  (* rinitialisation du monde (lors d'une surcharge, appeler la mthode hrite) *)
  method reset: unit

end



(* la classe world *)
class ['obj] world = 
object (self)

  (* le plateau *)
  val mutable cells: 'obj cell_t array array = [|[||]|]
  val invalid_cell = (new invalid_cell)#coerce
  method cell (x,y) = 
    try cells.(x).(y) 
    with _ -> invalid_cell

  (* obtention des cases correspondant  une liste de positions *)
  method cell_list = 
    let rec aux acc = function
      | (x,y)::q -> 
	  (* on retire les cases invalides au besoin *)
	  (match (try Some cells.(x).(y) with _ -> None) with
	     | Some c -> aux (c::acc) q
	     | None -> aux acc q
	  )
      | [] -> acc
    in
      aux []


  (* itration d'une fonction sur l'ensemble des cases *)
  method private iter f = 
    Array.iter (Array.iter f) cells

  (* homologue de List.fold_left *)
  method private fold_left f v = 
    Array.fold_left (Array.fold_left f) v cells


  (* sortie  l'cran *)
  method print =
    for i=(-1) to Array.length cells do
      for j=(-1) to Array.length cells.(0) do
	(self#cell (j,i))#print
      done;
      print_newline()
    done
    

  (* dplacement d'un objet 
   * cette mthode lve l'exception Cell.Busy si la case est occupe
   *)
  method deplace (o: 'obj) (dir: dir_t) =
    let cell' = self#cell (walk o#pos dir) in
      cell'#accept o dir


  (* enregistrement d'un nouveau gnrateur d'vnements *)
  val mutable generators = []
  method register_generator (g: unit -> Events.event list) = generators <- g::generators

  (* enregistrement d'un nouvel couteur d'vnements *)
  val mutable listeners = []
  method register_listener (l: Events.event -> unit) = listeners <- l::listeners

  (* enregistrement d'un nouvel actionnaire :-) *)
  val mutable reacters = []
  method register_reacter (r: unit -> unit) = reacters <- r::reacters

  (* tat de la partie : gagne/perdue/? *)
  val mutable won = None
  method won   = won <- Some true
  method lost  = match won with 
    | None -> won <- Some false
    | _ -> ()

  (* rinitialisation *)
  method reset = won <- None

  (* jouer un tour *)
  method next = 
    (* on gnre dans un premier temps tous les vnements, 
     * et on rcupre la liste des vnements globaux
     *)
    let events = List.fold_left (fun acc generator -> List.rev_append (generator()) acc) [] generators in
      (* a-t-on gagn, perdu, ou doit-on continuer *)
      match won with
	| Some true  -> Won
	| Some _     -> Lost
	| _ ->
	    (* on diffuse cette liste  tous les listeners *)
	    List.iter (fun e -> List.iter (fun listener -> listener e) listeners) events;
	    (* puis on fait ragir ces derniers *)
	    List.iter (fun reacter -> reacter()) reacters;    
	    (* on peut alors continuer *)
	    Continue


  (* retrait des objets "inutiles" *)
  method clean =
    self#iter (fun cell -> cell#clean)



end
	  

