/*
 *
 * Copyright (C) 2002 Aurelien Moreau, Antoni Gutsens Diaz, Benjamin Leveque,
 * 	 Damien Pous, David Roger, Lionel Vaux, Mathieu Martin, Nicolas Bernard,
 * 	 Samuel Thibault, Simon Capern, Thomas Gazagnaire
 * 
 * This file is part of Rigobot.
 * 
 * Rigobot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Ribogot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ribogot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/* 
 * the major part of this code has been written by Robert Bridson
 * 
 * textures.c: helper functions for managing textures from OCAML/LABLGL.
 * This provides routines to load in a mipmap texture from a PPM image on disk
 * (returning an id), to bind a loaded texture, and to destroy a texture.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <jpeglib.h>
#include <caml/memory.h>
#include <caml/misc.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include "textures.h"
#define CTOI(C) (*(int*)&C)


/* attempt to load a PPM file from the given filename */
GLubyte *readPPM (const char *fname, GLint *glw, GLint *glh)
{
  FILE *in;
  int c, w, h, maxc, n;
  GLubyte *img;

  in = fopen (fname, "rb");
  if (in == NULL)
    return NULL;

  /* make sure we're getting raw rgb data */
  if (fgetc(in) != 'P') { fclose(in); return NULL; }
  if (fgetc(in) != '6') { fclose(in); return NULL; }

  for (;;) {
    while (isspace(c = fgetc(in)));
    if (c == EOF) { fclose(in); return NULL; }
    else if (c == '#') while ((c = fgetc(in)) != '\n');
    else {
      ungetc (c, in);
      break;
    }
  }
  if (fscanf (in, "%d", &w) < 1) { fclose(in); return NULL; }

  for (;;) {
    while (isspace(c = fgetc(in)));
    if (c == EOF) { fclose(in); return NULL; }
    else if (c == '#') while ((c = fgetc(in)) != '\n');
    else {
      ungetc (c, in);
      break;
    }
  }
  if (fscanf (in, "%d", &h) < 1) { fclose(in); return NULL; }

  for (;;) {
    while (isspace(c = fgetc(in)));
    if (c == EOF) { fclose(in); return NULL; }
    else if (c == '#') while ((c = fgetc(in)) != '\n');
    else {
      ungetc (c, in);
      break;
    }
  }
  if (fscanf (in, "%d", &maxc) < 1) { fclose(in); return NULL; }
  if (maxc <=0 || maxc >255) { fclose(in); return NULL; }

  /* single character of whitespace */
  fgetc(in);

  /* allocate space for image */
  img = (GLubyte*) malloc (w*h*3);
  if (img == NULL) { fclose(in); return NULL; }

  /* read it in */
  n = fread (img, 1, w*h*3, in);
  fclose (in);
  if (n != w*h*3) { free(img); return NULL; }

  /* let us know the size of the image */
  *glw = w;
  *glh = h;

  return img;
}


GLubyte *readBMP (const char *File, GLint *Width, GLint *Height,GLint *c,GLenum *f,GLenum *t){
  unsigned char	   Header[0x36];
  int x,size;
  GLubyte             *Data;
  GLint		DataPos,DataSize,w,h;
  GLint			Components;
  GLenum		Format,Type;
  FILE		        *fichier;
  //printf("init\n");
  fichier = fopen(File,"rb");
  //printf("open\n");
  if (!fichier)
    return NULL;
  if (fread(Header,1,0x36,fichier)!=0x36) return NULL;
  //printf("fread\n");
  if (Header[0]!='B' || Header[1]!='M')	return NULL	;
  //printf("header0\n");
  if (CTOI(Header[0x1E])!=0)		return NULL		;
  //printf("header1\n");
  if (CTOI(Header[0x1C])!=24){/*printf("%d\n",	CTOI(Header[0x1C]));	*/return NULL		;}
  //printf("header2\n");
  DataPos			= CTOI(Header[0x0A]);
  //printf("header2\n");
  DataSize		        = CTOI(Header[0x22]);
  //printf("header2\n");
  w			= CTOI(Header[0x12]);
  h		          = CTOI(Header[0x16]);
  //printf("fin header %d %d\n",w,h);
  size=(h)*(w);
  Type = GL_UNSIGNED_BYTE;
  Format = GL_RGB;
  Components = 3;
  if (DataSize==0)
    DataSize=size*Components;
  if (DataPos==0)
    DataPos=0x36;
  //printf("fin header\n");
  fseek(fichier,DataPos,0);
  //printf("fseek\n");
  Data = (GLubyte*) malloc (size*3*sizeof(GLubyte));
  //printf("malloc\n");
  if (!Data)  return NULL;

  if (fread(Data,1,DataSize,fichier)!=DataSize)
    {
      free( Data);
      fclose(fichier);
      return NULL;
    }
  //  printf("close\n");
  fclose(fichier);

  // printf("inverse\n");
  for ( x=0;x<size;x++)
    { unsigned char t;
      t=Data[x*3];
      Data[x*3]=Data[x*3+2];
      Data[x*3+2]=t;
    }
  //printf("valeur\n");
  *c=Components;
  *f=Format;
  *t=Type;
  *Width=w;
  *Height=h;
  //printf("return\n");
    return   Data;
}







void DecodeJPG(struct jpeg_decompress_struct* cinfo, tImageJPG *pImageData)
{
  int i;
  unsigned char** rowPtr;
  int rowsRead;
	jpeg_read_header(cinfo, TRUE);
	
	jpeg_start_decompress(cinfo);
	pImageData->rowSpan = cinfo->image_width * cinfo->num_components;
	pImageData->sizeX   = cinfo->image_width;
	pImageData->sizeY   = cinfo->image_height;
	
	pImageData->data =(unsigned char*)malloc(sizeof( unsigned char)*pImageData->rowSpan * pImageData->sizeY);
		
	rowPtr =(unsigned char **) malloc(sizeof( unsigned char *)*pImageData->sizeY);
	for (i = 0; i < pImageData->sizeY; i++)
		rowPtr[i] = &(pImageData->data[i*pImageData->rowSpan]);
	rowsRead = 0;
	while (cinfo->output_scanline < cinfo->output_height) 
	{
	  rowsRead += jpeg_read_scanlines(cinfo, &rowPtr[rowsRead], cinfo->output_height - rowsRead);
	}
	
	free( rowPtr);
	jpeg_finish_decompress(cinfo);
}


GLubyte *readJPG(const char *filename, GLint *W, GLint *H)
{
	struct jpeg_decompress_struct cinfo;
	tImageJPG *pImageData = NULL;
	FILE *pFile;
	GLubyte * Data;
	struct jpeg_error_mgr jerr;
	
	if((pFile = fopen(filename, "rb")) == NULL) 
	{
		printf( "Unable to load JPG File!\n");
		return NULL;
	}
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, pFile);
	pImageData = (tImageJPG*)malloc(sizeof(tImageJPG));
	DecodeJPG(&cinfo, pImageData);
	jpeg_destroy_decompress(&cinfo);
	fclose(pFile);
	*W=pImageData->sizeX ;
	*H=pImageData->sizeY ;
	Data=pImageData->data;
	free(pImageData);
	return Data;
}












/* Attempt to open a PPM image from the given filename, build a 2d mipmap
 * texture from it, and create the id of a texture object representing it.
 * The id is a Nativeint in OCAML - not a regular int.
 */
value loadPPMTexture (value filename)
{
  CAMLparam1(filename);

  GLubyte *img;
  GLsizei w, h;
  GLuint texname;

  img = readPPM (String_val(filename), &w, &h);
  if (img == NULL) {
    return copy_nativeint(0);
  }

  /* get an id */
  glGenTextures (1, &texname);
  glBindTexture (GL_TEXTURE_2D, texname);

  /* create the mipmap texture */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  gluBuild2DMipmaps (GL_TEXTURE_2D, GL_RGB, w, h, GL_RGB, GL_UNSIGNED_BYTE,img);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
  /* free storage */
  free (img);

  CAMLreturn (copy_nativeint(texname));
}



value loadBMPTexture (value File){
  CAMLparam1(File);

  GLubyte *Data;
  GLsizei     Width,Height;
  GLint			Components;
  GLenum		Format,Type;
  GLuint texname;

  Data=readBMP( String_val(File),&Width,&Height,& Components,&Format,&Type  );
  if (Data == NULL) {
    return copy_nativeint(0);
  }
   /* get an id */
  glGenTextures (1, &texname);
  glBindTexture (GL_TEXTURE_2D, texname);

  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  gluBuild2DMipmaps (GL_TEXTURE_2D, Format, Width, Height,Format,Type,Data);
  /*
  glTexImage2D(GL_TEXTURE_2D,0,3,Width,Height,0,GL_RGB,GL_UNSIGNED_BYTE,Data);
*/
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
  
  free (Data);

  CAMLreturn (copy_nativeint(texname));
}



value loadJPGTexture (value File){
  CAMLparam1(File);

  GLubyte *Data;
  GLsizei     Width,Height;
  GLint			Components;
  GLenum		Format,Type;
  GLuint texname;

  Data=readJPG( String_val(File),&Width,&Height);
  if (Data == NULL) {
    return copy_nativeint(0);
  }
   /* get an id */
  glGenTextures (1, &texname);
  glBindTexture (GL_TEXTURE_2D, texname);

  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  gluBuild2DMipmaps (GL_TEXTURE_2D, GL_RGB, Width, Height,GL_RGB,GL_UNSIGNED_BYTE,Data);
  /*
  glTexImage2D(GL_TEXTURE_2D,0,3,Width,Height,0,GL_RGB,GL_UNSIGNED_BYTE,Data);
*/
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
  
  free (Data);

  CAMLreturn (copy_nativeint(texname));
}
/* bind a 2d texture */
value bind2dTexture (value texname)
{
  CAMLparam1(texname);
  GLuint id = (GLuint) Nativeint_val(texname);
  glBindTexture (GL_TEXTURE_2D, (GLuint) Nativeint_val(texname));
  CAMLreturn(Val_unit);
}


/* delete a bound texture */
value deleteOneTexture (value texname)
{
  CAMLparam1(texname);
  GLuint tex = (GLuint) Nativeint_val(texname);
  glDeleteTextures (1, &tex);
  CAMLreturn(Val_unit);
}

