/*
 * Copyright (C) 2002 Aurelien Moreau, Antoni Gutsens Diaz, Benjamin Leveque,
 * 	 Damien Pous, David Roger, Lionel Vaux, Mathieu Martin, Nicolas Bernard,
 * 	 Samuel Thibault, Simon Capern, Thomas Gazagnaire
 * 
 * This file is part of Rigobot.
 * 
 * Rigobot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Ribogot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ribogot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/* language/parser.mly 
 *
 *   Yacc sources for Rigobot parser. 
 *   Valid for both interactive and non-interactive mode.
 *
 */

%{
open Syntax;;

type parse_error =
  | Empty_seq
  | Unknown_seq

(* exception Parser of parse_error *)

(* let parse_error e = raise (Parser e) *)

%}

%token STEP STEP_B STEP_L STEP_R STEPTO GOTO CMDSEP JMP UTURN ROT_R ROT_L 
%token ACT TAKE PUT MARK UNMARK DISP
%token BREAK STOP REP TIMES END
%token LPAR RPAR MUSHR SEMISEMI
%token PLUS MINUS MULT DIV LPAR RPAR
%token LET CHANGE WHEN DO
%token <string> VARNAME
%token <int> INT
%token IF ELSE DO WHILE 
%token EQ LE GE LT GT NEQ

%left PLUS MINUS        /* lowest precedence */
%left MULT DIV          /* medium precedence */
%nonassoc UMINUS        /* highest precedence */

%type <Syntax.ast_t> start
%start  start




%%


/************** start **************/

/* Parse une squence de commandes se terminant par ;; */
/* Le ;; provoque l'valuation */
start:
| seq SEMISEMI {$1}
/* | SEMISEMI { parse_error Empty_seq} */
/* | error SEMISEMI { parse_error Unknown_seq } */
;

/************** seq **************/

/* Une squence de commandes */
seq:
| cmd {$1}
| cmd CMDSEP seq {Seq ($1,$3)}
;

/************** cmd **************/

/* Une commande est ... */
cmd:
  /* une action du robot : */
| action {$1}
  /* une dclaration : */
| declare_var { Nothing }
| declare_fun { Nothing }
| declare_handler { Nothing }
  /* la rptition d'une squence : */
| REP INT TIMES seq END {RepeatN ($2,$4)}     
  /* un brachement conditionnel : */
| ite {$1}
  /* une boucle conditionnelle : */
    /*| while {$1} */
;

/************** action **************/

/* Les actions du robot */
action: 
| mouv   { $1 }
| MARK   { Action SetMark}
| UNMARK { Action UnsetMark }
| TAKE   { Action Take }
| PUT    { Action Put }
| ACT    { Action Act }
;

/************** mouv **************/

/* Un mouvement du robot : */
mouv:
  /* un dplacement simple : */
| depl { Action $1 }
  /* un dplacement d'un certain nombre de cases */
| depl INT { let n = $2 in 
	       if (n=0) 
	       then Nothing 
	       else if (n=1) 
	       then Action $1
	       else let d = $1 in
		 match d with 
		   | Step -> RepeatN (n,(Action Step))
		   | other_step -> Seq (
		       Action other_step,
		       RepeatN (n-1, Action Step)
		     )
	   } 
  /* une rotation : */
| ROT_L { Action Rotate_Left }
| ROT_R { Action Rotate_Right }
| UTURN { Action UTurn }
  /* un saut en deux temps : */
| JMP { Seq((Action PreJump),(Action Jump)) }
;

/************** depl **************/

depl: 
| STEP   { Step }
| STEP_B { Step_Back }
| STEP_L { Step_Left }
| STEP_R { Step_Right }
;

/************** declare_* **************/

declare_var:
| LET VARNAME EQ arithm { NameSpace.declare_var $2 $4 }
| CHANGE VARNAME EQ arithm { NameSpace.change_var $2 $4 }
;

declare_fun:
| LET VARNAME args EQ seq END { NameSpace.declare_fun $2 $3 $5 }
;

declare_handler:
| WHEN VARNAME args DO seq END { NameSpace.declare_handler $2 $3 $5 }
;

/************** arithm **************/

arithm:
| value { $1 }
| LPAR arithm RPAR { $2 }
| arithm PLUS arithm { arithm_plus $1 $3 }
| arithm MINUS arithm { arithm_minus $1 $3 }
| arithm MULT arithm { arithm_mult $1 $3 }
| arithm DIV arithm { arithm_div $1 $3 }
| MINUS arithm %prec UMINUS { arithm_opposite $2 }
;

/************** value **************/

value:
| INT { Int ($1) }
| VARNAME { Name ($1) }
;

/************** ite **************/

ite:
| IF cond seq ELSE seq END { IfThenElse($2,$3,$5)}
| IF cond seq END { IfThenElse($2,$3,Nothing) } 
;

/************** args **************/

args:
| LPAR arglist RPAR {$2}
| LPAR RPAR {[]}
;

/************** arglist **************/
 
arglist:
| VARNAME {[$1]}
| VARNAME arglist { $1::$2 }
;

/************** cond **************/

cond: 
| arithm EQ arithm { Eq($1,$3) }
| arithm NEQ arithm { Neq($1,$3) }
| arithm LE arithm { Le($1,$3) }
| arithm GE arithm { Ge($1,$3) }
| arithm LT arithm { Lt($1,$3) }
| arithm GT arithm { Gt($1,$3) }
;

