(*
 *
 * Copyright (C) 2002 Aurelien Moreau, Antoni Gutsens Diaz, Benjamin Leveque,
 * 	 Damien Pous, David Roger, Lionel Vaux, Mathieu Martin, Nicolas Bernard,
 * 	 Samuel Thibault, Simon Capern, Thomas Gazagnaire
 * 
 * This file is part of Rigobot.
 * 
 * Rigobot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Ribogot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ribogot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)
open HeightField
open StdLabels

let eucli a b =
  let aux = floor (a/.b) in
    (aux, a-.aux*.b)

type look_t = int;;

type 'a option = None | Some of 'a;;

type square_t = {triangleList:GlList.t; frameList:GlList.t;};;

(* roption_t : render option => les options du rendu de la carte *)
type roption_t = [`Texture];;
(* doption_t : display option => option d'affichage de la carte *)
type doption_t = [`Grid|`LimitRayon];;

type patern_t = bool array array;;

let color c =
  let r= c/16/16/16/16 in
  let g= (c-r*16*16*16*16)/16/16 in
  let b= ((c-r*16*16*16*16)-g*16*16) in
    [|r;g;b|]
;;

let image_width = 64
and image_height = 64
let pi = acos (-1.0)


let make_image () =
  let src = 
    Graphic_image.array_of_image (Bmp.load "texture.bmp" [])
  in
  let image_width,image_height= Array.length src, Array.length src.(0) in
  let image =
    GlPix.create `ubyte ~height:image_width ~width:image_height ~format:`rgb in
  let raw = GlPix.to_raw image
  and pos = GlPix.raw_pos image in
  for i = 0 to image_width - 1 do
    for j = 0 to image_height - 1 do
      Raw.sets raw ~pos:(pos ~x:j ~y:i) (color (src.(j).(i)));
    done;
  done;
  image;;
(*
let make_image () =
  let image =
    GlPix.create `ubyte ~height:image_height ~width:image_width ~format:`rgb in
  let raw = GlPix.to_raw image
  and pos = GlPix.raw_pos image in
  for i = 0 to image_width - 1 do
    let ti = 2.0 *. pi *. float i /. float image_width in
    for j = 0 to image_height - 1 do
      let tj = 2.0 *. pi *. float j /. float image_height in
      Raw.sets raw ~pos:(pos ~x:j ~y:i)
	(Array.map ~f:(fun x -> truncate (127.0 *. (1.0 +. x)))
	   [|sin ti; cos (2.0 *. ti); cos (ti +. tj)|]);
	
      done;
  done;
  image
*)




(* interface de l'objet glMap3d *)
class type t=
object  
  method init                  : int -> float -> unit

  method get_dim               : unit -> (int*int)
  method get_squareDim         : unit -> (float*float)

  method draw_all              : unit -> unit
  method draw                  : (int*int) -> (float*float) -> unit
  method draw_ext              : (int*int) -> (float*float) -> patern_t -> unit

  method get_z                 : (float*float) -> float

  method enable_roption        : roption_t -> unit
  method disable_roption       : roption_t -> unit

  method enable_doption        : doption_t -> unit
  method disable_doption       : doption_t -> unit
end;;

class glMap3d hField (lField:look_t) : t =
object(self)
  
  val heightField = hField
  val lookField = lField

  val mutable dimx = 0
  val mutable dimy = 0

  val mutable optTexture = false
				  
  val mutable optGrid = true
  val mutable optLimitRayon = true

  val mutable squareDimx = 0
  val mutable squareDimy = 0
  val mutable squareMap = None
			    
  (* init lance la triangularisation  partir du heightField et enregistre la map dans openGl
     ratio reprsente le rapport (nb point de heightField)/(case de jeu square)
     aprox est un coef d'aproximation des points (donc reprsente une triangularisation parfaite)
  *)

  method init ratio aprox = 
    squareDimx<-ratio;
    squareDimy<-ratio;
    dimx<-(HeightField.get_dimX heightField)/ratio;
    dimy<-(HeightField.get_dimY heightField)/ratio;
    
    let image=make_image ()in
      GlTex.parameter ~target:`texture_2d (`min_filter(`nearest));
      GlTex.parameter ~target:`texture_2d (`mag_filter(`nearest));
      GlTex.image2d image;

    let tab =Array.make_matrix dimx dimy None in
    squareMap<- Some(tab);
      for i=0 to dimx-2 do
	for j=0 to dimy-2 do 
	    tab.(i).(j)<-self#create_square (i,j)
	done
      done
	      
  method get_dim () =
    (dimx,dimy)

  method get_squareDim () =
    (float_of_int squareDimx),(float_of_int squareDimy)

  method draw_all () = match squareMap with
    | None -> ()
    | Some(tab) -> begin
	for i=0 to dimx-1 do
	  for j=0 to dimy-1 do
	    self#draw_square(i,j) (true & optGrid)
	  done
	done
      end

  method draw (ox,oy) (vx,vy) =
    let r = int_of_float (sqrt (vx**2.+.vy**2.))+1 in
      for i=(-r) to r do
	for j=(-r) to r do
	  if self#is_draw (ox+i,oy+j) (ox,oy) (vx,vy) then 
	    self#draw_square (ox+i,oy+j) (i<2 & i>(-2) & j<2 & j>(-2) & optGrid)
	  else
	    ()
	done
      done

  method draw_ext (ox,oy) (vx,vy) patern =
    let r = int_of_float (sqrt (vx**2.+.vy**2.)) in
      for i=0 to 2*r do
	for j=0 to 2*r do
	  let test = try patern.(i).(j) with _ -> true in
	    if self#is_draw (ox-r+i,oy-r+j) (ox,oy) (vx,vy) & test then
	      self#draw_square (ox-r+i,oy-r+j) (i<(r+10) & i>(r-10) & j>(r-10) & j<(r+10) & optGrid)
	    else
	      ()
	done
      done

  method get_z (fx,fy) =
    let x = fx*.(float_of_int squareDimx) in
    let y = fy*.(float_of_int squareDimy) in
    let ix = floor x in
    let iy = floor y in
    let iix=int_of_float ix in
    let iiy=int_of_float iy in
    let dx = x-.ix in
    let dy = y-.iy in
    let h1=HeightField.get heightField (iix,iiy) in
    let h3=HeightField.get heightField (iix+1,iiy+1) in
      if dx>dy then
	let h2=HeightField.get heightField (iix+1,iiy) in
	  dx*.(h2-.h1)+.dy*.(h3-.h2)+.h1
      else
	let h2=HeightField.get heightField (iix,iiy+1)in
	  dx*.(h3-.h2)+.dy*.(h2-.h1)+.h1

  method enable_roption opt = match opt with
    |`Texture -> optTexture<-true

  method disable_roption opt = match opt with
    |`Texture -> optTexture<-false

  method enable_doption opt = match opt with
    |`Grid       -> optGrid <- true
    |`LimitRayon -> optLimitRayon <-true
      
  method disable_doption opt = match opt with
    |`Grid       -> optGrid <- false
    |`LimitRayon -> optLimitRayon <- false      

  method private draw_square (x,y) printSquare = match squareMap with
    | None -> ()
    | Some(tabSquare) -> begin
	try begin
	  match tabSquare.(x).(y) with
	    | None -> ()
	    | Some(square) ->
		GlList.call square.triangleList;
		if printSquare then (GlList.call square.frameList);
	end with _ -> ()
      end

  method private create_square (x,y) = match squareMap with
    | None -> None
    | Some(tabSquare) -> begin
	let minx=x*squareDimx in
	let maxx=minx+squareDimx in
	let miny=y*squareDimy in
	let maxy=miny+squareDimy in
	let fminy= float_of_int miny in
	let fmaxy= float_of_int maxy in
	let fminx= float_of_int minx in
	let fmaxx= float_of_int maxx in
	let tl= GlList.create `compile in
	  for i = minx to maxx-1 do
	    for j= miny to maxy-1 do
	      let fi=float_of_int i in
	      let fj=float_of_int j in
	      let z1,z2,z3,z4=
		(HeightField.get heightField (i,j)),
		(HeightField.get heightField (i+1,j)),
		(HeightField.get heightField (i,j+1)),
		(HeightField.get heightField (i+1,j+1)) 
	      in
	      let c1,c2,c3,c4=
		let colorize z =
		  (* if z>6. then (z*.0.03,0.,0.) else (0.,0., 0.5+.z*.0.03) *)
		  (z*.0.05,z*.0.05,z*.0.05)
		in
		  colorize z1,
		  colorize z2,
		  colorize z3,
		  colorize z4
	      in
		GlDraw.begins `triangles;
		GlDraw.color c1;
		GlTex.coord2 (0.,0.);
		GlDraw.vertex3 (fi,fj,z1);

		GlDraw.color c4;
		GlTex.coord2 (1.,1.);	
		GlDraw.vertex3 (fi+.1.,fj+.1.,z4);

		GlDraw.color c2;
		GlTex.coord2 (1.,0.);
		GlDraw.vertex3 (fi+.1.,fj,z2);

		GlDraw.ends ();
		GlDraw.begins `triangles;

		GlDraw.color c1;
		GlTex.coord2 (0.,0.);
		GlDraw.vertex3 (fi,fj,z1);

		GlDraw.color c4;
		GlTex.coord2 (1.,1.);
		GlDraw.vertex3 (fi+.1.,fj+.1.,z4);

		GlDraw.color c3;
		GlTex.coord2 (0.,1.);
		GlDraw.vertex3 (fi,fj+.1.,z3);

		GlDraw.ends ();
	    done
	  done;
	  GlList.ends ();	
	  let fl = GlList.create `compile in
	    GlDraw.color (0.,0.,0.);
	    for i = minx to maxx-1 do
	      let fi= float_of_int i in
	      let z1,z2,z3,z4 =
		HeightField.get heightField (i,miny)+.0.1,
		HeightField.get heightField (i+1,miny)+.0.1,
		HeightField.get heightField (i,maxy)+.0.1,
		HeightField.get heightField (i+1,maxy)+.0.1
	      in
		GlDraw.begins `lines;
		GlDraw.vertex3 (fi, fminy, z1);
		GlDraw.vertex3 (fi+.1.,fminy, z2);
		GlDraw.ends ();
		GlDraw.begins `lines;
		GlDraw.vertex3 (fi, fmaxy, z3);
		GlDraw.vertex3 (fi+.1.,fmaxy, z4);
		GlDraw.ends ();
	    done;
	    for i = miny to maxy-1 do
	      let fi = float_of_int i in
	      let z1,z2,z3,z4 = 
		HeightField.get heightField (minx,i) +.0.1,
		HeightField.get heightField (minx,i+1) +.0.1,
		HeightField.get heightField (maxx,i) +.0.1,
		HeightField.get heightField (maxx,i+1) +.0.1 
	      in
		GlDraw.begins `lines;
		GlDraw.vertex3 (fminx, fi, z1);
		GlDraw.vertex3 (fminx, fi+.1., z2);
		GlDraw.ends ();
		GlDraw.begins `lines;
		GlDraw.vertex3 (fmaxx, fi, z3);
		GlDraw.vertex3 (fmaxx, fi+.1., z4);
		GlDraw.ends ();
	    done;
	    GlList.ends ();
	    Some({triangleList=tl;
		  frameList=fl})
      end
	
  method private is_draw (x,y) (ox,oy) (vx,vy) =
    if optLimitRayon=false then
      true
    else
      let dx = float_of_int (x-ox) in
      let dy = float_of_int (y-oy) in
	(dx**2.+.dy**2.) < (vx**2.+.vy**2.) & (dx*.vx+.dy*.vy)>(-0.5) 
end;;
    
let glMap3d_from_file file =
  let height= HeightField.load_bmpHeight file in
    new glMap3d height 0
