(*
 *
 * Copyright (C) 2002 Aurelien Moreau, Antoni Gutsens Diaz, Benjamin Leveque,
 * 	 Damien Pous, David Roger, Lionel Vaux, Mathieu Martin, Nicolas Bernard,
 * 	 Samuel Thibault, Simon Capern, Thomas Gazagnaire
 * 
 * This file is part of Rigobot.
 * 
 * Rigobot is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Ribogot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Ribogot; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)
type t = float array array

let middle_height = 
  5.
;;

let floor_height = 
  0.
;;

let ceil_height = 
  40.
;;

let color_to_level c =
  let r= c/16/16/16/16 in
  let g= (c-r*16*16*16*16)/16/16 in
  let b= ((c-r*16*16*16*16)-g*16*16) in
    ((float_of_int (256-(r+g+b)/3))/.256.)*.(ceil_height-.floor_height)+.floor_height
;;


let heightField_of_floatMatrix (x:float array array)= 
  x
;;

exception HeightField_error of string;;



let load_bmpHeight file =
  let image = 
    try
      Bmp.load file [] 
    with _ -> raise (HeightField_error "Can't find or bad bmp file") 
  in
  let map_int= Graphic_image.array_of_image image in
  let dx,dy = Array.length map_int, Array.length map_int.(0) in
  let ret_tab= Array.make_matrix dx dy 0.0 in
    for i=0 to dx-1 do
      for j=0 to dy-1 do
	ret_tab.(i).(j)<- color_to_level map_int.(i).(j)
      done
    done;
    ret_tab
;;


type relief_t = 

    Hill of float*float
      (* hauteur pente *)
  | Plateau of float*float*(int*int)
      (* hauteur pente dimension *)
  | Valley of float*float*(int*int)
      (* profondeur pente dimension *)
  | Lack of float*float
      (* profondeur pente *)
;;

let gen_height x y = 
  if x<>0 & y<>0 then
    Array.make_matrix x y middle_height
  else 
    raise (HeightField_error "Can't generate zero heightField")
;;

let get_dimX field = 
  Array.length field
;;

let get_dimY field = 
  Array.length field.(0)
;;

let get field (x,y) =
  field.(x).(y)
;;

let pi = 
  3.14159265359
;;




let height_modif field (x,y) h_tmp p (dx,dy) =
  let h,comp,sign= if h_tmp > 0. then h_tmp,max,1. else (-.h_tmp),min,-1. in 
  let supp = int_of_float (h/.p) in
  let startx,endx = min x (dx+x), max x (dx+x) in
  let starty,endy = min y (dy+y), max y (dy+y) in	
    (* le plateau *)
    for i= startx+supp to endx-supp do
      for j= starty+supp to endy-supp do
	(try
	   field.(i).(j) <-  comp (middle_height+.h_tmp) field.(i).(j)
	 with _ -> ())
      done
    done;
    (* les coins *)
    for i=0 to supp do
      for j=0 to supp do
	let level = middle_height+.
		      ((1.+.cos ((sqrt ((float_of_int i)**2.+.(float_of_int j)**2.))/.(h/.p)*.pi)) *.h)/.2.*.sign in
	  if (sqrt ((float_of_int i)**2.+.(float_of_int j)**2.) < (float_of_int supp)) then 
	    begin	      
	      (try
		 field.(startx+supp-i).(starty+supp-j)<- comp level field.(startx+supp-i).(starty+supp-j)
	       with _ -> ());		  
	      (try
		 field.(startx+supp-i).(endy-supp+j)<- comp level field.(startx+supp-i).(endy-supp+j)
	       with _ -> ());
	      (try
		 field.(endx-supp+i).(starty+supp-j) <- comp level field.(endx-supp+i).(starty+supp-j)
	       with _ -> ());
	      (try
		 field.(endx-supp+i).(endy-supp+j) <- comp level field.(endx-supp+i).(endy-supp+j)
	       with _ -> ())
	    end
	  else
	    ()
      done
    done;
    
    (* les rebords *)
    for j=0 to supp do
      let level = middle_height+.
		    ((1.+.cos ((float_of_int (abs j))/.(h/.p)*.pi))*.h)/.2.*.sign in
	for i=startx+supp to (endx-supp) do
	  (try
	     field.(i).(starty+supp-j)<- comp level field.(i).(starty+supp-j)
	   with _ -> ());		  
	  (try
	     field.(i).(endy-supp+j)<- comp level field.(i).(endy-supp+j)
	   with _ -> ())
	done;
	for i=(starty+supp) to (endy-supp) do
	  (try
	     field.(startx+supp-j).(i)<- comp level field.(startx+supp-j).(i)
	   with _ -> (););		  
	  (try
	     field.(endx-supp+j).(i)<- comp level field.(endx-supp+j).(i)
	   with _ -> (););	    
	done
    done;
    
    field
    

let put_relief field relief (x,y)= match relief with
  | Hill(h_sign,p) -> 
      let h= if h_sign>0. then h_sign else (-.h_sign) in
      let coef = int_of_float (h/.p) in
	height_modif field (x-coef,y-coef) h p (2*coef, 2*coef);
	field
		
  | Plateau(h_sign,p,(dx,dy)) -> 
      let h= if h_sign>0. then h_sign else (-.h_sign) in
	if (float_of_int dx)> (h/.p*.2.) & (float_of_int dy)> (h/.p*.2.) then
	  height_modif field (x,y) h p (dx, dy) 
	else   
	  raise (HeightField_error "Bad plateau value");
	field

  | Valley(h_sign,p,(dx,dy)) -> 
      let h= if h_sign>0. then h_sign else (-.h_sign) in
	if (float_of_int dx)> (h/.p*.2.) & (float_of_int dy)> (h/.p*.2.) then
	  height_modif field (x,y) (-.h) p (dx, dy) 
	else  
	  raise (HeightField_error "Bad valley value");
	field

  | Lack(h_sign,p) -> 
      let h= if h_sign>0. then h_sign else (-.h_sign) in
      let coef = int_of_float (h/.p) in
	height_modif field (x-coef,y-coef) (-.h) p (2*coef, 2*coef);
	field

;;
